@@ -6,39 +6,45 @@ import os
 import re
 import time
 import traceback
+import platform
 from datetime import datetime, timezone
 
 import aiofiles
-import aiohttp
+import httpx
 from tqdm import tqdm
 
-__version__ = "2.0.5"
+__version__ = "2.0.7"
 CONFIG_PATH = os.path.join(
     os.path.dirname(os.path.abspath(__file__)), ".canvassyncer.json"
 )
 PAGES_PER_TIME = 8
 
 
-class AsyncDownloader:
-    def __init__(self, sess, sem, config):
-        self.sess: aiohttp.ClientSession = sess
-        self.sem = sem
-        self.config = config
+class AsyncSemClient:
+    def __init__(self, connectionCount, token, proxies):
+        self.sem = asyncio.Semaphore(connectionCount)
+        self.client = httpx.AsyncClient(
+            timeout=5,
+            headers={"Authorization": f"Bearer {token}"},
+            proxies=proxies,
+            transport=httpx.AsyncHTTPTransport(retries=3),
+        )
 
     async def downloadOne(self, src, dst):
         async with self.sem:
-            async with self.sess.get(src, proxy=self.config.get("proxies")) as res:
-                if res.status != 200:
+            async with self.client.stream("GET", src) as res:
+                if res.status_code != 200:
                     return self.failures.append(f"{src} => {dst}")
+                num_bytes_downloaded = res.num_bytes_downloaded
                 async with aiofiles.open(dst, "+wb") as f:
-                    while True:
-                        chunk = await res.content.read(1024 * 4)
-                        if not chunk:
-                            break
+                    async for chunk in res.aiter_bytes():
                         await f.write(chunk)
-                        self.tqdm.update(len(chunk))
+                        self.tqdm.update(
+                            res.num_bytes_downloaded - num_bytes_downloaded
+                        )
+                        num_bytes_downloaded = res.num_bytes_downloaded
 
-    async def start(self, infos, totalSize=0):
+    async def downloadMany(self, infos, totalSize=0):
         self.tqdm = tqdm(total=totalSize, unit="B", unit_scale=True)
         self.failures = []
         await asyncio.gather(
@@ -50,46 +56,66 @@ class AsyncDownloader:
             for text in self.failures:
                 print(text)
 
+    async def json(self, *args, **kwargs):
+        async with self.sem:
+            resp = await self.client.get(*args, **kwargs)
+        res = resp.json()
+        if (
+            kwargs.get("checkError") == True
+            and isinstance(res, dict)
+            and res.get("errors")
+        ):
+            errMsg = res["errors"][0].get("message", "unknown error.")
+            print(f"\nError: {errMsg}")
+            exit(1)
+        return res
+
+    async def head(self, *args, **kwargs):
+        async with self.sem:
+            resp = await self.client.head(*args, **kwargs)
+        return resp.headers
+
+    async def aclose(self):
+        await self.client.aclose()
+
 
 class CanvasSyncer:
     def __init__(self, config):
-        self.confirmAll = config["y"]
         self.config = config
-        self.sess = aiohttp.ClientSession()
-        self.sem = asyncio.Semaphore(config["connection_count"])
+        self.client = AsyncSemClient(
+            config["connection_count"], config["token"], config.get("proxies")
+        )
         self.downloadSize = 0
         self.laterDownloadSize = 0
         self.courseCode = {}
-        self.baseurl = self.config["canvasURL"] + "/api/v1"
+        self.baseUrl = self.config["canvasURL"] + "/api/v1"
         self.downloadDir = self.config["downloadDir"]
         self.newInfo = []
         self.newFiles = []
         self.laterFiles = []
         self.laterInfo = []
         self.skipfiles = []
-        self.downloader = AsyncDownloader(self.sess, self.sem, self.config)
+        self.totalFileCount = 0
         if not os.path.exists(self.downloadDir):
             os.mkdir(self.downloadDir)
 
-    async def close(self):
-        await self.sess.close()
+    async def aclose(self):
+        await self.client.aclose()
 
-    def setSessionArgs(self, **kwargs):
-        kwargs["timeout"] = kwargs.get("timeout", 10)
-        kwargs["headers"] = kwargs.get("headers", {})
-        kwargs["headers"]["Authorization"] = f"Bearer {self.config['token']}"
-        kwargs["proxy"] = self.config.get("proxies")
-        return kwargs
-
-    async def sessGetJson(self, *args, **kwargs):
-        async with self.sem:
-            async with self.sess.get(*args, **self.setSessionArgs(**kwargs)) as resp:
-                return await resp.json()
-
-    async def sessHead(self, *args, **kwargs):
-        async with self.sem:
-            async with self.sess.head(*args, **self.setSessionArgs(**kwargs)) as resp:
-                return resp.headers
+    async def dictFromPages(self, helperFunc, *args, **kwargs):
+        res = {}
+        page = 1
+        endOfPage = False
+        while not endOfPage:
+            pageRes = await asyncio.gather(
+                *[helperFunc(page + i, *args, **kwargs) for i in range(PAGES_PER_TIME)]
+            )
+            for item in pageRes:
+                if not item:
+                    endOfPage = True
+                res.update(item)
+            page += PAGES_PER_TIME
+        return res
 
     def prepareLocalFiles(self, courseID, folders):
         localFiles = []
@@ -109,10 +135,10 @@ class CanvasSyncer:
             ]
         return localFiles
 
-    async def getCourseFoldersWithIDHelper(self, courseID, page):
+    async def getCourseFoldersWithIDHelper(self, page, courseID):
         res = {}
-        url = f"{self.baseurl}/courses/{courseID}/folders?page={page}"
-        folders = await self.sessGetJson(url)
+        url = f"{self.baseUrl}/courses/{courseID}/folders?page={page}"
+        folders = await self.client.json(url)
         for folder in folders:
             if folder["full_name"].startswith("course files"):
                 folder["full_name"] = folder["full_name"][len("course files") :]
@@ -122,28 +148,10 @@ class CanvasSyncer:
             res[folder["id"]] = re.sub(r"[\\\:\*\?\"\<\>\|]", "_", res[folder["id"]])
         return res
 
-    async def getCourseFoldersWithID(self, courseID):
-        folders = {}
-        page = 1
-        endOfPage = False
-        while not endOfPage:
-            pageRes = await asyncio.gather(
-                *[
-                    self.getCourseFoldersWithIDHelper(courseID, page + i)
-                    for i in range(PAGES_PER_TIME)
-                ]
-            )
-            for item in pageRes:
-                if not item:
-                    endOfPage = True
-                folders.update(item)
-            page += PAGES_PER_TIME
-        return folders
-
-    async def getCourseFilesHelper(self, courseID, page, folders):
+    async def getCourseFilesHelper(self, page, courseID, folders):
         files = {}
-        url = f"{self.baseurl}/courses/{courseID}/files?page={page}"
-        canvasFiles = await self.sessGetJson(url)
+        url = f"{self.baseUrl}/courses/{courseID}/files?page={page}"
+        canvasFiles = await self.client.json(url)
         if not canvasFiles or isinstance(canvasFiles, dict):
             return files
         for f in canvasFiles:
@@ -158,32 +166,14 @@ class CanvasSyncer:
         return files
 
     async def getCourseFiles(self, courseID):
-        files = {}
-        page = 1
-        folders = await self.getCourseFoldersWithID(courseID)
-        endOfPage = False
-        while not endOfPage:
-            pageRes = await asyncio.gather(
-                *[
-                    self.getCourseFilesHelper(courseID, page + i, folders)
-                    for i in range(PAGES_PER_TIME)
-                ]
-            )
-            for item in pageRes:
-                if not item:
-                    endOfPage = True
-                files.update(item)
-            page += PAGES_PER_TIME
+        folders = await self.dictFromPages(self.getCourseFoldersWithIDHelper, courseID)
+        files = await self.dictFromPages(self.getCourseFilesHelper, courseID, folders)
         return folders, files
 
     async def getCourseIdByCourseCodeHelper(self, page, lowerCourseCodes):
         res = {}
-        url = f"{self.baseurl}/courses?page={page}"
-        courses = await self.sessGetJson(url)
-        if isinstance(courses, dict) and courses.get("errors"):
-            errMsg = courses["errors"][0].get("message", "unknown error.")
-            print(f"\nError: {errMsg}")
-            exit(1)
+        url = f"{self.baseUrl}/courses?page={page}"
+        courses = await self.client.json(url, checkError=True)
         if not courses:
             return res
         for course in courses:
@@ -193,28 +183,17 @@ class CanvasSyncer:
         return res
 
     async def getCourseIdByCourseCode(self):
-        page = 1
         lowerCourseCodes = [s.lower() for s in self.config["courseCodes"]]
-        endOfPage = False
-        while not endOfPage:
-            pageRes = await asyncio.gather(
-                *[
-                    self.getCourseIdByCourseCodeHelper(page + i, lowerCourseCodes)
-                    for i in range(PAGES_PER_TIME)
-                ]
-            )
-            for item in pageRes:
-                if not item:
-                    endOfPage = True
-                self.courseCode.update(item)
-            page += PAGES_PER_TIME
+        self.courseCode = await self.dictFromPages(
+            self.getCourseIdByCourseCodeHelper, lowerCourseCodes
+        )
 
     async def getCourseCodeByCourseIDHelper(self, courseID):
-        url = f"{self.baseurl}/courses/{courseID}"
-        sessRes = await self.sessGetJson(url)
-        if sessRes.get("course_code") is None:
+        url = f"{self.baseUrl}/courses/{courseID}"
+        clientRes = await self.client.json(url)
+        if clientRes.get("course_code") is None:
             return
-        self.courseCode[courseID] = sessRes["course_code"]
+        self.courseCode[courseID] = clientRes["course_code"]
 
     async def getCourseCodeByCourseID(self):
         await asyncio.gather(
@@ -246,7 +225,7 @@ class CanvasSyncer:
         path = path.replace("\\", "/").replace("//", "/")
         if fileName in localFiles and fileModifiedTimeStamp <= os.path.getctime(path):
             return
-        response = await self.sessHead(fileUrl)
+        response = await self.client.head(fileUrl)
         fileSize = int(response.get("content-length", 0))
         if fileName in localFiles:
             self.laterDownloadSize += fileSize
@@ -256,9 +235,6 @@ class CanvasSyncer:
             )
             return
         if fileSize > self.config["filesizeThresh"] * 1000000:
-            print(
-                f"{self.courseCode[courseID]}{fileName} ({round(fileSize / 1000000, 2)}MB) ignored, too large."
-            )
             aiofiles.open(path, "w").close()
             self.skipfiles.append(
                 f"{self.courseCode[courseID]}{fileName} ({round(fileSize / 1000000, 2)}MB)"
@@ -272,6 +248,7 @@ class CanvasSyncer:
 
     async def getCourseTaskInfo(self, courseID):
         folders, files = await self.getCourseFiles(courseID)
+        self.totalFileCount += len(files)
         localFiles = self.prepareLocalFiles(courseID, folders)
         await asyncio.gather(
             *[
@@ -301,7 +278,7 @@ class CanvasSyncer:
         print("These file(s) have later version on canvas:")
         for s in self.laterInfo:
             print(s)
-        isDownload = "Y" if self.confirmAll else input("Update all?(Y/n) ")
+        isDownload = "Y" if self.config["y"] else input("Update all?(Y/n) ")
         if isDownload in ["n", "N"]:
             return
         print(f"Start to download {len(self.laterInfo)} file(s)!")
@@ -313,7 +290,9 @@ class CanvasSyncer:
                     ntpath.dirname(path),
                     f"{localCreatedTimeStamp}_{ntpath.basename(path)}",
                 )
-                if not os.path.exists(newPath):
+                if self.config["no_keep_older_version"]:
+                    os.remove(path)
+                elif not os.path.exists(newPath):
                     os.rename(path, newPath)
                 else:
                     path = os.path.join(
@@ -336,11 +315,12 @@ class CanvasSyncer:
                 for courseID in self.courseCode.keys()
             ]
         )
+        print(f"Get {self.totalFileCount} files!")
         if not self.newFiles and not self.laterFiles:
             return print("All local files are synced!")
         self.checkNewFiles()
         self.checkLaterFiles()
-        await self.downloader.start(
+        await self.client.downloadMany(
             self.newFiles + self.laterFiles, self.downloadSize + self.laterDownloadSize
         )
 
@@ -352,9 +332,10 @@ def initConfig():
     elif os.path.exists("./canvassyncer.json"):
         oldConfig = json.load(open("./canvassyncer.json"))
     print("Generating new config file...")
-    url = input("Canvas url(Default: https://umjicanvas.com):").strip()
+    prevu = oldConfig.get("canvasURL", "") if oldConfig else "https://umjicanvas.com"
+    url = input("Canvas url(Defuault: " + prevu + "):").strip()
     if not url:
-        url = "https://umjicanvas.com"
+        url = prevu
     tipStr = f"(Default: {oldConfig.get('token', '')})" if oldConfig else ""
     token = input(f"Canvas access token{tipStr}:").strip()
     if not token:
@@ -410,72 +391,98 @@ def initConfig():
     )
 
 
+def getConfig():
+    parser = argparse.ArgumentParser(description="A Simple Canvas File Syncer")
+    parser.add_argument("-r", help="recreate config file", action="store_true")
+    parser.add_argument("-y", help="confirm all prompts", action="store_true")
+    parser.add_argument(
+        "--no-subfolder",
+        help="do not create a course code named subfolder when synchronizing files",
+        action="store_true",
+    )
+    parser.add_argument(
+        "-p", "--path", help="appoint config file path", default=CONFIG_PATH
+    )
+    parser.add_argument(
+        "-c",
+        "--connection",
+        help="max connection count with server",
+        default=16,
+        type=int,
+    )
+    parser.add_argument("-x", "--proxy", help="download proxy", default=None)
+    parser.add_argument("-V", "--version", action="version", version=__version__)
+    parser.add_argument(
+        "-d", "--debug", help="show debug information", action="store_true"
+    )
+    parser.add_argument(
+        "--no-keep-older-version",
+        help="do not keep older version",
+        action="store_true",
+    )
+    args = parser.parse_args()
+    configPath = args.path
+    if args.r or not os.path.exists(configPath):
+        if not os.path.exists(configPath):
+            print("Config file not exist, creating...")
+        try:
+            initConfig()
+        except Exception as e:
+            print(f"\nError: {e.__class__.__name__}. Failed to create config file.")
+            if args.debug:
+                print(traceback.format_exc())
+            exit(1)
+        if args.r:
+            return
+    config = json.load(open(configPath, "r", encoding="UTF-8"))
+    config["y"] = args.y
+    config["proxies"] = args.proxy
+    config["no_subfolder"] = args.no_subfolder
+    config["connection_count"] = args.connection
+    config["no_keep_older_version"] = args.no_keep_older_version
+    config["debug"] = args.debug
+    return config
+
+
 async def sync():
-    Syncer, args = None, None
+    syncer = None
     try:
-        parser = argparse.ArgumentParser(description="A Simple Canvas File Syncer")
-        parser.add_argument("-r", help="recreate config file", action="store_true")
-        parser.add_argument("-y", help="confirm all prompts", action="store_true")
-        parser.add_argument(
-            "--no-subfolder",
-            help="do not create a course code named subfolder when synchronizing files",
-            action="store_true",
-        )
-        parser.add_argument(
-            "-p", "--path", help="appoint config file path", default=CONFIG_PATH
-        )
-        parser.add_argument(
-            "-c",
-            "--connection",
-            help="max connection count with server",
-            default=16,
-            type=int,
-        )
-        parser.add_argument("-x", "--proxy", help="download proxy", default=None)
-        parser.add_argument("-V", "--version", action="version", version=__version__)
-        parser.add_argument(
-            "-d", "--debug", help="show debug information", action="store_true"
-        )
-        args = parser.parse_args()
-        configPath = args.path
-        if args.r or not os.path.exists(configPath):
-            if not os.path.exists(configPath):
-                print("Config file not exist, creating...")
+        config = getConfig()
+        while True:
             try:
-                initConfig()
-            except Exception as e:
-                print(f"\nError: {e.__class__.__name__}. Failed to create config file.")
-                if args.debug:
-                    print(traceback.format_exc())
-                exit(1)
-            if args.r:
-                return
-        config = json.load(open(configPath, "r", encoding="UTF-8"))
-        config["y"] = args.y
-        config["proxies"] = args.proxy
-        config["no_subfolder"] = args.no_subfolder
-        config["connection_count"] = args.connection
-        Syncer = CanvasSyncer(config)
-        await Syncer.sync()
-    except aiohttp.ServerDisconnectedError as e:
-        print("Server disconnected error, try to reduce connection count using -c")
-        exit(1)
+                syncer = CanvasSyncer(config)
+                await syncer.sync()
+                break
+            except httpx.ConnectError as e:
+                if config["connection_count"] == 2:
+                    raise e
+                config["connection_count"] //= 2
+                print(
+                    "Server connect error, reducing connection count "
+                    f"to {config['connection_count']} and retrying..."
+                )
     except KeyboardInterrupt as e:
-        raise (e)
+        raise e
     except Exception as e:
         errorName = e.__class__.__name__
-        print(f"Unexpected error: {errorName}. Please check your network and token! Or use -d for detailed information.")
-        if args.debug:
+        print(
+            f"Unexpected error: {errorName}. Please check your network and token!",
+            end="",
+        )
+        print("" if config["debug"] else " Or use -d for detailed information.")
+        if config["debug"]:
             print(traceback.format_exc())
     finally:
-        await Syncer.close()
+        if syncer:
+            await syncer.aclose()
 
 
 def run():
-    asyncio.set_event_loop(asyncio.new_event_loop())
     try:
-        asyncio.get_event_loop().run_until_complete(sync())
-    except KeyboardInterrupt as e:
+        if platform.system() == "Windows":
+            asyncio.set_event_loop_policy(asyncio.WindowsSelectorEventLoopPolicy())
+        asyncio.run(sync())
+    except KeyboardInterrupt:
         print("\nOperation cancelled by user, exiting...")
         exit(1)
 
